/* ARCADEmini assembler (unix)
 * version 1.6 date 2004-11-02
 * part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
 * Copyright (C) 2003-2004 1stein <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 */

%{
#include <stdio.h>
#include <string.h>

#include "id_list.h"
#include "node.h"
#include "parse.h"
#include "parse_.h"
#include "scan.h"
#include "global.h"
#include "debug.h"

int string_caller; //start condition that called string
int comment_caller; //start condition that called comment
%}

IDENT [A-Za-z_][A-Za-z_0-9]*

NUM_DEZ [1-9][0-9]*
NUM_OCT 0[0-7]*
NUM_HEX 0[Xx][0-9A-Fa-f]*
NUM_BIN 0[Bb][01]*
NUM [+-]?({NUM_DEZ}|{NUM_OCT}|{NUM_HEX}|{NUM_BIN})

CHAR_NORM [^'"\a\b\f\n\r\t\v\0]
CHAR_ESC "\\"['"\\aAbBfFnNrRtTvV]
CHAR_HEX "\\"[xX][0-9A-Fa-f][0-9A-Fa-f]?
CHAR_OCT "\\"[0-7][0-7]?[0-7]?
CHAR ({CHAR_NORM}|{CHAR_ESC}|{CHAR_HEX}|{CHAR_OCT})

%s CMD
%x CMNT

%%

<CMD>"REGISTER" { debug1( "TOKEN: CMD_REGISTER, TEXT: \"%s\"", yytext ); column += 8; BEGIN( INITIAL ); return CMD_REGISTER; }
<CMD>"NEXT_REG" { debug1( "TOKEN: CMD_NEXT_REG, TEXT: \"%s\"", yytext ); column += 8; BEGIN( INITIAL ); return CMD_NEXT_REG; }
<CMD>"NEXT_SEC" { debug1( "TOKEN: CMD_NEXT_SEC, TEXT: \"%s\"", yytext ); column += 8; BEGIN( INITIAL ); return CMD_NEXT_SEC; }
<CMD>"DB" { debug1( "TOKEN: CMD_DB, TEXT: \"%s\"", yytext ); column += 2; BEGIN( INITIAL ); return CMD_DB; }
<CMD>"DW" { debug1( "TOKEN: CMD_DW, TEXT: \"%s\"", yytext ); column += 2; BEGIN( INITIAL ); return CMD_DW; }
<CMD>"DD" { debug1( "TOKEN: CMD_DD, TEXT: \"%s\"", yytext ); column += 2; BEGIN( INITIAL ); return CMD_DD; }
<CMD>"JMP_NEXT_SEC" { debug1( "TOKEN: CMD_JMP_NEXT_SEC, TEXT: \"%s\"", yytext ); column += 12; BEGIN( INITIAL ); return CMD_JMP_NEXT_SEC; }
<CMD>"JMP_SEC" { debug1( "TOKEN: CMD_JMP_SEC, TEXT: \"%s\"", yytext ); column += 7; BEGIN( INITIAL ); return CMD_JMP_SEC; }
<CMD>"JMP_OFS" { debug1( "TOKEN: CMD_JMP_OFS, TEXT: \"%s\"", yytext ); column += 7; BEGIN( INITIAL ); return CMD_JMP_OFS; }
<CMD>"JMP_OFS_SEC" { debug1( "TOKEN: CMD_JMP_OFS_SEC, TEXT: \"%s\"", yytext ); column += 11; BEGIN( INITIAL ); return CMD_JMP_OFS_SEC; }
<CMD>"JMP_IND" { debug1( "TOKEN: CMD_JMP_IND, TEXT: \"%s\"", yytext ); column += 7; BEGIN( INITIAL ); return CMD_JMP_IND; }
<CMD>"JMP_EQ" { debug1( "TOKEN: CMD_JMP_EQ, TEXT: \"%s\"", yytext ); column += 6; BEGIN( INITIAL ); return CMD_JMP_EQ; }
<CMD>"JMP_NEQ" { debug1( "TOKEN: CMD_JMP_NEQ, TEXT: \"%s\"", yytext ); column += 7; BEGIN( INITIAL ); return CMD_JMP_NEQ; }
<CMD>"JMP_GT" { debug1( "TOKEN: CMD_JMP_GT, TEXT: \"%s\"", yytext ); column += 6; BEGIN( INITIAL ); return CMD_JMP_GT; }
<CMD>"JMP_NGT" { debug1( "TOKEN: CMD_JMP_NGT, TEXT: \"%s\"", yytext ); column += 7; BEGIN( INITIAL ); return CMD_JMP_NGT; }
<CMD>"JMP_LT" { debug1( "TOKEN: CMD_JMP_LT, TEXT: \"%s\"", yytext ); column += 6; BEGIN( INITIAL ); return CMD_JMP_LT; }
<CMD>"JMP_NLT" { debug1( "TOKEN: CMD_JMP_NLT, TEXT: \"%s\"", yytext ); column += 7; BEGIN( INITIAL ); return CMD_JMP_NLT; }
<CMD>"JMP_GTS" { debug1( "TOKEN: CMD_JMP_GTS, TEXT: \"%s\"", yytext ); column += 7; BEGIN( INITIAL ); return CMD_JMP_GTS; }
<CMD>"JMP_NGTS" { debug1( "TOKEN: CMD_JMP_NGTS, TEXT: \"%s\"", yytext ); column += 8; BEGIN( INITIAL ); return CMD_JMP_NGTS; }
<CMD>"JMP_LTS" { debug1( "TOKEN: CMD_JMP_LTS, TEXT: \"%s\"", yytext ); column += 7; BEGIN( INITIAL ); return CMD_JMP_LTS; }
<CMD>"JMP_NLTS" { debug1( "TOKEN: CMD_JMP_NLTS, TEXT: \"%s\"", yytext ); column += 8; BEGIN( INITIAL ); return CMD_JMP_NLTS; }
<CMD>"JMP_AND_Z" { debug1( "TOKEN: CMD_JMP_AND_Z, TEXT: \"%s\"", yytext ); column += 9; BEGIN( INITIAL ); return CMD_JMP_AND_Z; }
<CMD>"JMP_AND_NZ" { debug1( "TOKEN: CMD_JMP_AND_NZ, TEXT: \"%s\"", yytext ); column += 10; BEGIN( INITIAL ); return CMD_JMP_AND_NZ; }
<CMD>"JMP_OR_M" { debug1( "TOKEN: CMD_JMP_OR_M, TEXT: \"%s\"", yytext ); column += 8; BEGIN( INITIAL ); return CMD_JMP_OR_M; }
<CMD>"JMP_OR_NM" { debug1( "TOKEN: CMD_JMP_OR_NM, TEXT: \"%s\"", yytext ); column += 9; BEGIN( INITIAL ); return CMD_JMP_OR_NM; }
<CMD>"MOV" { debug1( "TOKEN: CMD_MOV, TEXT: \"%s\"", yytext ); column += 3; BEGIN( INITIAL ); return CMD_MOV; }
<CMD>"MOV2" { debug1( "TOKEN: CMD_MOV2, TEXT: \"%s\"", yytext ); column += 4; BEGIN( INITIAL ); return CMD_MOV2; }
<CMD>"MOV4" { debug1( "TOKEN: CMD_MOV2, TEXT: \"%s\"", yytext ); column += 4; BEGIN( INITIAL ); return CMD_MOV4; }
<CMD>"RD_I" { debug1( "TOKEN: CMD_RD_I, TEXT: \"%s\"", yytext ); column += 4; BEGIN( INITIAL ); return CMD_RD_I; }
<CMD>"WR_I" { debug1( "TOKEN: CMD_WR_I, TEXT: \"%s\"", yytext ); column += 4; BEGIN( INITIAL ); return CMD_WR_I; }
<CMD>"RD_S" { debug1( "TOKEN: CMD_RD_S, TEXT: \"%s\"", yytext ); column += 4; BEGIN( INITIAL ); return CMD_RD_S; }
<CMD>"RD_CF" { debug1( "TOKEN: CMD_RD_CF, TEXT: \"%s\"", yytext ); column += 5; BEGIN( INITIAL ); return CMD_RD_CF; }
<CMD>"ADD" { debug1( "TOKEN: CMD_ADD, TEXT: \"%s\"", yytext ); column += 3; BEGIN( INITIAL ); return CMD_ADD; }
<CMD>"ADDC" { debug1( "TOKEN: CMD_ADDC, TEXT: \"%s\"", yytext ); column += 4; BEGIN( INITIAL ); return CMD_ADDC; }
<CMD>"SUB" { debug1( "TOKEN: CMD_SUB, TEXT: \"%s\"", yytext ); column += 3; BEGIN( INITIAL ); return CMD_SUB; }
<CMD>"SUBB" { debug1( "TOKEN: CMD_SUBB, TEXT: \"%s\"", yytext ); column += 4; BEGIN( INITIAL ); return CMD_SUBB; }
<CMD>"MUL" { debug1( "TOKEN: CMD_MUL, TEXT: \"%s\"", yytext ); column += 3; BEGIN( INITIAL ); return CMD_MUL; }
<CMD>"MULS" { debug1( "TOKEN: CMD_MULS, TEXT: \"%s\"", yytext ); column += 4; BEGIN( INITIAL ); return CMD_MULS; }
<CMD>"DIV" { debug1( "TOKEN: CMD_DIV, TEXT: \"%s\"", yytext ); column += 3; BEGIN( INITIAL ); return CMD_DIV; }
<CMD>"DIVS" { debug1( "TOKEN: CMD_DIVS, TEXT: \"%s\"", yytext ); column += 4; BEGIN( INITIAL ); return CMD_DIVS; }
<CMD>"AND" { debug1( "TOKEN: CMD_AND, TEXT: \"%s\"", yytext ); column += 3; BEGIN( INITIAL ); return CMD_AND; }
<CMD>"OR" { debug1( "TOKEN: CMD_OR, TEXT: \"%s\"", yytext ); column += 2; BEGIN( INITIAL ); return CMD_OR; }
<CMD>"XOR" { debug1( "TOKEN: CMD_XOR, TEXT: \"%s\"", yytext ); column += 3; BEGIN( INITIAL ); return CMD_XOR; }
<CMD>"SHL" { debug1( "TOKEN: CMD_SHL, TEXT: \"%s\"", yytext ); column += 3; BEGIN( INITIAL ); return CMD_SHL; }
<CMD>"SHR" { debug1( "TOKEN: CMD_SHR, TEXT: \"%s\"", yytext ); column += 3; BEGIN( INITIAL ); return CMD_SHR; }
<CMD>"ROL" { debug1( "TOKEN: CMD_ROL, TEXT: \"%s\"", yytext ); column += 3; BEGIN( INITIAL ); return CMD_ROL; }
<CMD>"ROR" { debug1( "TOKEN: CMD_ROR, TEXT: \"%s\"", yytext ); column += 3; BEGIN( INITIAL ); return CMD_ROR; }
<CMD>"PUSH" { debug1( "TOKEN: CMD_PUSH, TEXT: \"%s\"", yytext ); column += 4; BEGIN( INITIAL ); return CMD_PUSH; }
<CMD>"PUSH_C" { debug1( "TOKEN: CMD_PUSH_C, TEXT: \"%s\"", yytext ); column += 6; BEGIN( INITIAL ); return CMD_PUSH_C; }
<CMD>"POP" { debug1( "TOKEN: CMD_POP, TEXT: \"%s\"", yytext ); column += 3; BEGIN( INITIAL ); return CMD_POP; }
<CMD>"CALL" { debug1( "TOKEN: CMD_CALL, TEXT: \"%s\"", yytext ); column += 4; BEGIN( INITIAL ); return CMD_CALL; }
<CMD>"CALL_IND" { debug1( "TOKEN: CMD_CALL_IND, TEXT: \"%s\"", yytext ); column += 8; BEGIN( INITIAL ); return CMD_CALL_IND; }
<CMD>"RET" { debug1( "TOKEN: CMD_RET, TEXT: \"%s\"", yytext ); column += 3; BEGIN( INITIAL ); return CMD_RET; }
<CMD>"RD_PIX" { debug1( "TOKEN: CMD_RD_PIX, TEXT: \"%s\"", yytext ); column += 6; BEGIN( INITIAL ); return CMD_RD_PIX; }
<CMD>"RD_OVP" { debug1( "TOKEN: CMD_RD_OVP, TEXT: \"%s\"", yytext ); column += 6; BEGIN( INITIAL ); return CMD_RD_OVP; }
<CMD>"WR_PIX" { debug1( "TOKEN: CMD_WR_PIX, TEXT: \"%s\"", yytext ); column += 6; BEGIN( INITIAL ); return CMD_WR_PIX; }
<CMD>"WR_OVP" { debug1( "TOKEN: CMD_WR_OVP, TEXT: \"%s\"", yytext ); column += 6; BEGIN( INITIAL ); return CMD_WR_OVP; }
<CMD>"WR_PIX_LINE" { debug1( "TOKEN: CMD_WR_PIX_LINE, TEXT: \"%s\"", yytext ); column += 11; BEGIN( INITIAL ); return CMD_WR_PIX_LINE; }
<CMD>"WR_OVP_LINE" { debug1( "TOKEN: CMD_WR_OVP_LINE, TEXT: \"%s\"", yytext ); column += 11; BEGIN( INITIAL ); return CMD_WR_OVP_LINE; }
<CMD>"WR_PIX_QUART" { debug1( "TOKEN: CMD_WR_PIX_QUART, TEXT: \"%s\"", yytext ); column += 12; BEGIN( INITIAL ); return CMD_WR_PIX_QUART; }
<CMD>"WR_OVP_QUART" { debug1( "TOKEN: CMD_WR_OVP_QUART, TEXT: \"%s\"", yytext ); column += 12; BEGIN( INITIAL ); return CMD_WR_OVP_QUART; }
<CMD>"WR_PIX_CLR" { debug1( "TOKEN: CMD_WR_PIX_CLR, TEXT: \"%s\"", yytext ); column += 10; BEGIN( INITIAL ); return CMD_WR_PIX_CLR; }
<CMD>"WR_OVP_CLR" { debug1( "TOKEN: CMD_WR_OVP_CLR, TEXT: \"%s\"", yytext ); column += 10; BEGIN( INITIAL ); return CMD_WR_OVP_CLR; }
<CMD>"WR_PIX_SET" { debug1( "TOKEN: CMD_WR_PIX_SET, TEXT: \"%s\"", yytext ); column += 10; BEGIN( INITIAL ); return CMD_WR_PIX_SET; }
<CMD>"WR_OVP_SET" { debug1( "TOKEN: CMD_WR_OVP_SET, TEXT: \"%s\"", yytext ); column += 10; BEGIN( INITIAL ); return CMD_WR_OVP_SET; }
<CMD>"WR_PIX_BW" { debug1( "TOKEN: CMD_WR_PIX_BW, TEXT: \"%s\"", yytext ); column += 9; BEGIN( INITIAL ); return CMD_WR_PIX_BW; }
<CMD>"WR_OVP_BW" { debug1( "TOKEN: CMD_WR_OVP_BW, TEXT: \"%s\"", yytext ); column += 9; BEGIN( INITIAL ); return CMD_WR_OVP_BW; }
<CMD>"WR_PIX_SEL" { debug1( "TOKEN: CMD_WR_PIX_SEL, TEXT: \"%s\"", yytext ); column += 10; BEGIN( INITIAL ); return CMD_WR_PIX_SEL; }
<CMD>"WR_OVP_SEL" { debug1( "TOKEN: CMD_WR_OVP_SEL, TEXT: \"%s\"", yytext ); column += 10; BEGIN( INITIAL ); return CMD_WR_OVP_SEL; }
<CMD>"CP_BLK_PP" { debug1( "TOKEN: CMD_CP_BLK_PP, TEXT: \"%s\"", yytext ); column += 6; BEGIN( INITIAL ); return CMD_CP_BLK_PP; }
<CMD>"CP_BLK_PO" { debug1( "TOKEN: CMD_CP_BLK_PO, TEXT: \"%s\"", yytext ); column += 6; BEGIN( INITIAL ); return CMD_CP_BLK_PO; }
<CMD>"CP_BLK_OP" { debug1( "TOKEN: CMD_CP_BLK_OP, TEXT: \"%s\"", yytext ); column += 6; BEGIN( INITIAL ); return CMD_CP_BLK_OP; }
<CMD>"CP_BLK_OO" { debug1( "TOKEN: CMD_CP_BLK_OO, TEXT: \"%s\"", yytext ); column += 6; BEGIN( INITIAL ); return CMD_CP_BLK_OO; }
<CMD>"CP_BLK_PP_TR" { debug1( "TOKEN: CMD_CP_BLK_PP_TR, TEXT: \"%s\"", yytext ); column += 9; BEGIN( INITIAL ); return CMD_CP_BLK_PP_TR; }
<CMD>"CP_BLK_PO_TR" { debug1( "TOKEN: CMD_CP_BLK_PO_TR, TEXT: \"%s\"", yytext ); column += 9; BEGIN( INITIAL ); return CMD_CP_BLK_PO_TR; }
<CMD>"CP_BLK_OP_TR" { debug1( "TOKEN: CMD_CP_BLK_OP_TR, TEXT: \"%s\"", yytext ); column += 9; BEGIN( INITIAL ); return CMD_CP_BLK_OP_TR; }
<CMD>"CP_BLK_OO_TR" { debug1( "TOKEN: CMD_CP_BLK_OO_TR, TEXT: \"%s\"", yytext ); column += 9; BEGIN( INITIAL ); return CMD_CP_BLK_OO_TR; }
<CMD>"DISPLAY" { debug1( "TOKEN: CMD_DISPLAY, TEXT: \"%s\"", yytext ); column += 7; BEGIN( INITIAL ); return CMD_DISPLAY; }
<CMD>"DISPLAY_CP" { debug1( "TOKEN: CMD_DISPLAY_CP, TEXT: \"%s\"", yytext ); column += 10; BEGIN( INITIAL ); return CMD_DISPLAY_CP; }
<CMD>"DELAY_DO" { debug1( "TOKEN: CMD_DELAY_DO, TEXT: \"%s\"", yytext ); column += 8; BEGIN( INITIAL ); return CMD_DELAY_DO; }
<CMD>"DELAY_START" { debug1( "TOKEN: CMD_DELAY_START, TEXT: \"%s\"", yytext ); column += 11; BEGIN( INITIAL ); return CMD_DELAY_START; }
<CMD>"DELAY_CHECK" { debug1( "TOKEN: CMD_DELAY_CHECK, TEXT: \"%s\"", yytext ); column += 11; BEGIN( INITIAL ); return CMD_DELAY_CHECK; }
<CMD>"DELAY_WAIT" { debug1( "TOKEN: CMD_DELAY_WAIT, TEXT: \"%s\"", yytext ); column += 10; BEGIN( INITIAL ); return CMD_DELAY_WAIT; }
<CMD>"GET_KEYS" { debug1( "TOKEN: CMD_GET_KEYS, TEXT: \"%s\"", yytext ); column += 8; BEGIN( INITIAL ); return CMD_GET_KEYS; }
<CMD>"GET_SNES0" { debug1( "TOKEN: CMD_GET_SNES0, TEXT: \"%s\"", yytext ); column += 9; BEGIN( INITIAL ); return CMD_GET_SNES0; }
<CMD>"GET_SNES1" { debug1( "TOKEN: CMD_GET_SNES1, TEXT: \"%s\"", yytext ); column += 9; BEGIN( INITIAL ); return CMD_GET_SNES1; }
<CMD>"GET_TIME" { debug1( "TOKEN: CMD_GET_TIME, TEXT: \"%s\"", yytext ); column += 8; BEGIN( INITIAL ); return CMD_GET_TIME; }
<CMD>"SET_TIME" { debug1( "TOKEN: CMD_SET_TIME, TEXT: \"%s\"", yytext ); column += 8; BEGIN( INITIAL ); return CMD_SET_TIME; }
<CMD>"RND" { debug1( "TOKEN: CMD_RND, TEXT: \"%s\"", yytext ); column += 3; BEGIN( INITIAL ); return CMD_RND; }
<CMD>"VER" { debug1( "TOKEN: CMD_VER, TEXT: \"%s\"", yytext ); column += 3; BEGIN( INITIAL ); return CMD_VER; }
<CMD>"NOP" { debug1( "TOKEN: CMD_NOP, TEXT: \"%s\"", yytext ); column += 3; BEGIN( INITIAL ); return CMD_NOP; }

{IDENT}":" { debug1( "TOKEN: LABEL_CON, TEXT: \"%s\"", yytext ); yylval.p_node = scan_ident_decl( yytext, 1, id_list_con ); column += strlen( yytext ); return LABEL_CON; }
{IDENT} { debug1( "TOKEN: IDENT_CON, TEXT: \"%s\"", yytext ); yylval.p_node = scan_ident_use( yytext, 0, id_list_con ); column += strlen( yytext ); return IDENT_CON; }
"$"{IDENT}":" { debug1( "TOKEN: LABEL_REG, TEXT: \"%s\"", yytext ); yylval.p_node = scan_ident_decl( yytext + 1, 1, id_list_reg ); column += strlen( yytext ); return LABEL_REG; }
"$"{IDENT} { debug1( "TOKEN: IDENT_REG, TEXT: \"%s\"", yytext ); yylval.p_node = scan_ident_use( yytext + 1, 0, id_list_reg ); column += strlen( yytext ); return IDENT_REG; }
"@"{IDENT}":" { debug1( "TOKEN: LABEL_ADDR, TEXT: \"%s\"", yytext ); yylval.p_node = scan_ident_decl( yytext + 1, 1, id_list_addr ); column += strlen( yytext ); return LABEL_ADDR; }
"@"{IDENT} { debug1( "TOKEN: IDENT_ADDR, TEXT: \"%s\"", yytext ); yylval.p_node = scan_ident_use( yytext + 1, 0, id_list_addr ); column += strlen( yytext ); return IDENT_ADDR; }

{NUM} { debug1( "TOKEN: NUMBER, TEXT: \"%s\"", yytext ); yylval.p_node = scan_con( yytext ); column += strlen( yytext ); return NUMBER; }
"\'"{CHAR}"\'" { debug1( "TOKEN: CHAR, TEXT: \"%s\"", yytext ); yylval.p_node = scan_char( yytext ); column += strlen( yytext ); return CHAR; }
"\'"{CHAR}{2,}"\'" { debug1( "TOKEN: STRING, TEXT: \"%s\"", yytext ); yylval.p_node = scan_string( yytext ); column += strlen( yytext ); return STRING; }
"\""{CHAR}*"\"" { debug1( "TOKEN: STRING, TEXT: \"%s\"", yytext ); yylval.p_node = scan_string( yytext ); column += strlen( yytext ); return STRING; }

"(" { debug1( "TOKEN: LBRA, TEXT: \"%s\"", yytext ); column++; return LBRA; }
")" { debug1( "TOKEN: RBRA, TEXT: \"%s\"", yytext ); column++; return RBRA; }
"~" { debug1( "TOKEN: NOT, TEXT: \"%s\"", yytext ); column++; return NOT; }
"*" { debug1( "TOKEN: MUL, TEXT: \"%s\"", yytext ); column++; return MUL; }
"/" { debug1( "TOKEN: DIV, TEXT: \"%s\"", yytext ); column++; return DIV; }
"%" { debug1( "TOKEN: MOD, TEXT: \"%s\"", yytext ); column++; return MOD; }
"+" { debug1( "TOKEN: ADD, TEXT: \"%s\"", yytext ); column++; return ADD; }
"-" { debug1( "TOKEN: SUB, TEXT: \"%s\"", yytext ); column++; return SUB; }
"<<" { debug1( "TOKEN: SHL, TEXT: \"%s\"", yytext );column += 2; return SHL; }
">>" { debug1( "TOKEN: SHR, TEXT: \"%s\"", yytext );column += 2; return SHR; }
"&" { debug1( "TOKEN: AND, TEXT: \"%s\"", yytext ); column++; return AND; }
"^" { debug1( "TOKEN: XOR, TEXT: \"%s\"", yytext ); column++; return XOR; }
"|" { debug1( "TOKEN: OR, TEXT: \"%s\"", yytext ); column++; return OR; }

".S" { debug1( "TOKEN: SEC, TEXT: \"%s\"", yytext ); column += 2; return SEC; }
".O" { debug1( "TOKEN: OFS, TEXT: \"%s\"", yytext ); column += 2; return OFS; }
"$" { debug1( "TOKEN: REG, TEXT: \"%s\"", yytext ); column++; return REG; }
"," { debug1( "TOKEN: COMMA, TEXT: \"%s\"", yytext ); column++; return COMMA; }
";" { debug1( "TOKEN: SEM, TEXT: \"%s\"", yytext ); column++; BEGIN( CMD ); return SEM; }

(\r\n|\r|\n) { debug1( "line %lu", line ); column = 1; line++; }
" "+ { column += strlen( yytext ); }
\t { column = ((column + 8) & 0xFFFFFFF8) + 1; }

("#"|"//").* { column += strlen( yytext ); }

<CMD>"/*" { comment_caller = CMD; BEGIN( CMNT ); column += 2; }
"/*" { comment_caller = INITIAL; BEGIN( CMNT ); column += 2; }
<CMNT>[^*\r\n]* { /* ignore anything that's not a '*' */ column += strlen( yytext ); }
<CMNT>"*"+[^*/\r\n]* { /* ignore '*'s not followed by '/'s */ column += strlen( yytext); }
<CMNT>(\r\n|\r|\n) { column = 1; line++; }
<CMNT>"*"+"/" { BEGIN( comment_caller ); column += strlen( yytext ); }

. { debug1( "TOKEN: UNKNOWN, TEXT: \"%s\"", yytext ); column++; return UNKNOWN; }
                                            
%%

int yywrap( )
{
  //no other file to process
  return 1;
}

//initialization function, called by main
void scan_init( FILE * p_input_file )
{
  //set input file
  yyrestart( p_input_file );
  //allow command as first token
  BEGIN( CMD );
}

//an identifier was declared
//remove the last rem_chr characters from identifier name
//put this identifier into the list
//create a node for identifier and return pointer to it
st_node * scan_ident_decl( char * p_name, unsigned int rem_chr, et_id_list_type type )
{
  char * p_terminate, chr_bak;
  st_id_list_entry * p_id;
  st_node * p_node;

  //add identifier to list

  //remove last rem_chr characters from identifier name
  p_terminate = p_name + strlen( p_name ) - rem_chr; //get position where to terminate the string
  chr_bak = *p_terminate; //backup character that will be overwritten with 0
  *p_terminate = 0; //terminate string at desired position

  debug2( "scan_ident_decl: p_name=%s type=%s", p_name, id_list_dbg_type2str( type ) );

  //add identifier to list
  p_id = id_list_add( p_name, type );
  if( p_id == NULL )
  {
    *p_terminate = chr_bak; //restore original identifier name
    return NULL;
  }

  //store position of declaration

  //identifier is not yet declared
  if( p_id->decl_column == 0 && p_id->decl_line == 0 )
  {
    //store current position as posititon of declaration
    p_id->decl_column = column;
    p_id->decl_line = line;
  }
  //identifier is already declared
  else
  {
    //print error message
    printf( "duplicate declaration of \"%s\" in column %lu in line %lu\n", p_name, column, line );
    error_cnt++;
  }

  //restore original identifier name
  *p_terminate = chr_bak;

  //create a node for this identifier
  switch( type )
  {
    case id_list_unused: p_node = NULL; break;
    case id_list_con: p_node = node_new_con_ident( p_id ); break;
    case id_list_reg: p_node = node_new_reg_ident( p_id ); break;
    case id_list_addr: p_node = node_new_addr_ident( p_id ); break;
    default: p_node = NULL;
  }
  //if this fails, we return the NULL we got (nothing need to be freed)

  //return pointer to node
  return p_node;
}

//an identifier was used
//remove the last rem_chr characters from identifier name
//put this identifier into the list
//create a node for identifier and return pointer to it
st_node * scan_ident_use( char * p_name, unsigned int rem_chr, et_id_list_type type )
{
  char * p_terminate, chr_bak;
  st_id_list_entry * p_id;
  st_node * p_node;

  //add identifier to list

  //remove last rem_chr characters from identifier name
  p_terminate = p_name + strlen( p_name ) - rem_chr; //get position where to terminate the string
  chr_bak = *p_terminate; //backup character that will be overwritten with 0
  *p_terminate = 0; //terminate string at desired position

  debug2( "scan_ident_use: p_name=%s type=%s", p_name, id_list_dbg_type2str( type ) );

  //add identifier to list
  p_id = id_list_add( p_name, type );
  if( p_id == NULL )
  {
    *p_terminate = chr_bak; //restore original identifier name
    return NULL;
  }

  //store position of first usage

  //identifier has not yet been used
  if( p_id->firstuse_column == 0 && p_id->firstuse_line == 0 )
  {
    //store current position as posititon of first usage
    p_id->firstuse_column = column;
    p_id->firstuse_line = line;
  }

  //restore original identifier name
  *p_terminate = chr_bak;

  //create a node for this identifier
  switch( type )
  {
    case id_list_unused: p_node = NULL; break;
    case id_list_con: p_node = node_new_con_ident( p_id ); break;
    case id_list_reg: p_node = node_new_reg_ident( p_id ); break;
    case id_list_addr: p_node = node_new_addr_ident( p_id ); break;
    default: p_node = NULL;
  }
  //if this fails, we return the NULL we got (nothing need to be freed)

  //return pointer to node
  return p_node;
}

//a constant was scanned
//create a node for constant and return pointer to it
st_node * scan_con( char * p_text )
{
  int neg, i;
  unsigned long val;
  st_node * p_node;

  //get negative sign
  neg = p_text[0] == '-';
  //jump over sign
  if( p_text[0] == '+' || p_text[0] == '-' )
    p_text++;

  //binary constant
  if( p_text[0] == '0' && (p_text[1] == 'B' || p_text[1] == 'b') )
  {
    val = 0;
    for( i = 2; p_text[i] != 0; i++ )
      if( p_text[i] >= '0' && p_text[i] <= '1' )
        val = val << 1 | (p_text[i] - '0');
  }
  //hex constant
  else if( p_text[0] == '0' && (p_text[1] == 'X' || p_text[1] == 'x') )
  {
    val = 0;
    for( i = 2; p_text[i] != 0; i++ )
      if( p_text[i] >= '0' && p_text[i] <= '9' )
        val = val << 4 | (p_text[i] - '0');
      else if( p_text[i] >= 'A' && p_text[i] <= 'F' )
        val = val << 4 | (p_text[i] - 'A' + 0xA);
      else if( p_text[i] >= 'a' && p_text[i] <= 'f' )
        val = val << 4 | (p_text[i] - 'a' + 0xA);
  }
  //octal constant
  else if( p_text[0] == '0' )
  {
    val = 0;
    for( i = 1; p_text[i] != 0; i++ )
      if( p_text[i] >= '0' && p_text[i] <= '7' )
        val = val << 3 | (p_text[i] - '0');
  }
  //decimal constant
  else
  {
    val = 0;
    for( i = 0; p_text[i] != 0; i++ )
      if( p_text[i] >= '0' && p_text[i] <= '9' )
        val = val * 10 + (p_text[i] - '0');
  }

  //process negative sign
  if( neg )
    val = (unsigned long)-(long)val;

  debug2( "scan_con: p_text=%s val=%lX", p_text, val );

  //create a node for this constant
  p_node = node_new_con( val );
  //if this fails, we return the NULL we got (nothing needs to be freed)

  //return pointer to node
  return p_node;
}

//convert a scanned character into its byte value
//returns the number of characters processed of p_text
unsigned int char2byte( char * p_text, unsigned char * p_byte_val )
{
  unsigned int i, len;
  unsigned char byte_val;

  //first character is a 0
  if( p_text[0] == 0 )
  {
    //no character available
    byte_val = 0;
    len = 0;
  }
  //first character is a backslash
  else if( p_text[0] == '\\' )
  {
    switch( p_text[1] )
    {
      case 0: byte_val = (unsigned char)'\\'; len = 1; break;
      case '\'': byte_val = (unsigned char)'\''; len = 2; break;
      case '\"': byte_val = (unsigned char)'\"'; len = 2; break;
      case '\\': byte_val = (unsigned char)'\\'; len = 2; break;
      case 'a': case 'A': byte_val = (unsigned char)'\a'; len = 2; break;
      case 'b': case 'B': byte_val = (unsigned char)'\b'; len = 2; break;
      case 'f': case 'F': byte_val = (unsigned char)'\f'; len = 2; break;
      case 'n': case 'N': byte_val = (unsigned char)'\n'; len = 2; break;
      case 'r': case 'R': byte_val = (unsigned char)'\r'; len = 2; break;
      case 't': case 'T': byte_val = (unsigned char)'\t'; len = 2; break;
      case 'v': case 'V': byte_val = (unsigned char)'\v'; len = 2; break;
      case 'x': case 'X': //hex constant
        byte_val = 0;
        for( i = 2; i < 4 && p_text[i] != 0; i++ )
          if( p_text[i] >= '0' && p_text[i] <= '9' )
            byte_val = byte_val << 4 | (p_text[i] - '0');
          else if( p_text[i] >= 'A' && p_text[i] <= 'F' )
            byte_val = byte_val << 4 | (p_text[i] - 'A' + 0xA);
          else if( p_text[i] >= 'a' && p_text[i] <= 'f' )
            byte_val = byte_val << 4 | (p_text[i] - 'a' + 0xA);
        len = i;
        break;
      case '0': case '1': case '2': case '3': case '4': case '5': case '6': case '7': //octal constant
        byte_val = 0;
        for( i = 1; i < 4 && p_text[i] != 0; i++ )
          if( p_text[i] >= '0' && p_text[i] <= '7' )
            byte_val = byte_val << 3 | (p_text[i] - '0');
            break;
        len = i;
      default: //use character behind backslash
        byte_val = p_text[1];
        len = 2;
    }
  }
  //first character is not a 0 and not a backslash
  else
  {
    //use first character
    byte_val = (unsigned char)p_text[0];
    len = 1;
  }

  //return byte value of character
  *p_byte_val = byte_val;

  //return number of charaxcters processed
  return len;
}

//a character was scanned
//create a node for character and return pointer to it
st_node * scan_char( char * p_text )
{
  unsigned int text_len;
  unsigned char byte_val;
  st_node * p_node;

  //get length of text
  text_len = strlen( p_text );

  //charcter is enclosed in single quotation marks
  if( text_len >= 2 && p_text[0] == '\'' && p_text[text_len - 1] == '\'' )
  {
    //remove single quotation marks
    p_text[text_len - 1] = 0;
    p_text++;
    text_len -= 2;
  }

  //get byte value of character
  char2byte( p_text, &byte_val );

  debug2( "scan_char: p_text=%s val=%X", p_text, byte_val );

  //create a node for this constant
  p_node = node_new_con( byte_val );
  //if this fails, we return the NULL we got (nothing needs to be freed)

  //return pointer to node
  return p_node;
}

//a string was scanned
//create a node for string and return pointer to it
st_node * scan_string( char * p_text )
{
  unsigned int text_len;
  unsigned int add_zero = 0; //if to add a zero byte behind end of string (0 or 1, not a boolean)
  unsigned int str_len, len, pos, i;
  unsigned char byte_val;
  st_node * p_node;

  //get length of text
  text_len = strlen( p_text );

  //charcter is enclosed in single quotation marks
  if( text_len >= 2 && p_text[0] == '\'' && p_text[text_len - 1] == '\'' )
  {
    //remove single quotation marks
    p_text[text_len - 1] = 0;
    p_text++;
    text_len -= 2;
  }
  //charcter is enclosed in double quotation marks
  else if( text_len >= 2 && p_text[0] == '\"' && p_text[text_len - 1] == '\"' )
  {
    //remove single quotation marks
    p_text[text_len - 1] = 0;
    p_text++;
    text_len -= 2;
    add_zero = 1;
  }

  //scan string to get length
  pos = 0;
  str_len = 0;
  for( ; ; )
  {
    len = char2byte( p_text + pos, &byte_val ); //scan next character
    if( len == 0 ) //end of string
      break;
    pos += len; //advance to next character in source text
    str_len++; //count number of characters in destination string
  }

  debug2( "scan_string: p_text=%s str_len=%d", p_text, str_len );

  //check maximum length
  if( str_len + add_zero > 255 )
  {
    printf( "WARNING: string too long (%u characters) in column %lu of line %lu, truncated after %u characters\n", str_len, column, line, 255 - add_zero );
    warning_cnt++;
    str_len = 255 - add_zero;
  }

  //create a node for this constant
  p_node = node_new_str( (unsigned char)(str_len + add_zero) );
  if( p_node == NULL )
    return NULL;

  //scan string and store data in node
  pos = 0;
  for( i = 0; i < str_len; i++ )
  {
    len = char2byte( p_text + pos, &p_node->data.str.data[i] ); //scan next character
    pos += len; //advance to next character in source text
  }

  //perhaps add zero byte
  if( add_zero )
    p_node->data.str.data[str_len] = 0;

  //return pointer to node
  return p_node;
}

